@@ -103,6 +103,8 @@ class LatexUnicodeTable:
         self.register(u'\n\n', b' \\par', encode=False)
         self.register(u'\n\n', b'\\par', encode=False)
         self.register(u' ', b'\\ ', encode=False)
+        self.register(u'\N{EM SPACE}', b'\\quad')
+        self.register(u'%', b'\\%')
         self.register(u'\N{EN DASH}', b'--')
         self.register(u'\N{EN DASH}', b'\\textendash')
         self.register(u'\N{EM DASH}', b'---')
@@ -111,9 +113,19 @@ class LatexUnicodeTable:
         self.register(u'\N{RIGHT SINGLE QUOTATION MARK}', b"'", decode=False)
         self.register(u'\N{LEFT DOUBLE QUOTATION MARK}', b'``')
         self.register(u'\N{RIGHT DOUBLE QUOTATION MARK}', b"''")
+        self.register(u'\N{DOUBLE LOW-9 QUOTATION MARK}', b'\\glqq')
         self.register(u'\N{DAGGER}', b'\\dag')
         self.register(u'\N{DOUBLE DAGGER}', b'\\ddag')
 
+        self.register(u'\\', b'\\textbackslash', encode=False)
+        self.register(u'\\', b'\\backslash', mode='math', encode=False)
+
+        self.register(u'\N{TILDE OPERATOR}', b'\\sim', mode='math')
+        self.register(u'\N{MODIFIER LETTER LOW TILDE}',
+                      b'\\texttildelow', package='textcomp')
+        self.register(u'\N{SMALL TILDE}', b'\\~{}')
+        self.register(u'~', b'\\textasciitilde')
+
         self.register(u'\N{BULLET}', b'\\bullet', mode='math')
         self.register(u'\N{BULLET}', b'\\textbullet', package='textcomp')
 
@@ -130,12 +142,14 @@ class LatexUnicodeTable:
         self.register(u'\N{SECTION SIGN}', b'\\S')
         self.register(u'\N{DIAERESIS}', b'\\"{}')
         self.register(u'\N{NOT SIGN}', b'\\neg')
+        self.register(u'\N{HYPHEN}', b'-', decode=False)
         self.register(u'\N{SOFT HYPHEN}', b'\\-')
         self.register(u'\N{MACRON}', b'\\={}')
 
         self.register(u'\N{DEGREE SIGN}', b'^\\circ', mode='math')
         self.register(u'\N{DEGREE SIGN}', b'\\textdegree', package='textcomp')
 
+        self.register(u'\N{MINUS SIGN}', b'-', mode='math')
         self.register(u'\N{PLUS-MINUS SIGN}', b'\\pm', mode='math')
         self.register(u'\N{PLUS-MINUS SIGN}', b'\\textpm', package='textcomp')
 
@@ -178,6 +192,8 @@ class LatexUnicodeTable:
         self.register(u'\N{LATIN CAPITAL LETTER A WITH TILDE}', b'\\~A')
         self.register(u'\N{LATIN CAPITAL LETTER A WITH DIAERESIS}', b'\\"A')
         self.register(u'\N{LATIN CAPITAL LETTER A WITH RING ABOVE}', b'\\AA')
+        self.register(u'\N{LATIN CAPITAL LETTER A WITH RING ABOVE}', b'\\r A',
+                      encode=False)
         self.register(u'\N{LATIN CAPITAL LETTER AE}', b'\\AE')
         self.register(u'\N{LATIN CAPITAL LETTER C WITH CEDILLA}', b'\\c C')
         self.register(u'\N{LATIN CAPITAL LETTER E WITH GRAVE}', b'\\`E')
@@ -207,6 +223,8 @@ class LatexUnicodeTable:
         self.register(u'\N{LATIN SMALL LETTER A WITH TILDE}', b'\\~a')
         self.register(u'\N{LATIN SMALL LETTER A WITH DIAERESIS}', b'\\"a')
         self.register(u'\N{LATIN SMALL LETTER A WITH RING ABOVE}', b'\\aa')
+        self.register(u'\N{LATIN SMALL LETTER A WITH RING ABOVE}', b'\\r a',
+                      encode=False)
         self.register(u'\N{LATIN SMALL LETTER AE}', b'\\ae')
         self.register(u'\N{LATIN SMALL LETTER C WITH CEDILLA}', b'\\c c')
         self.register(u'\N{LATIN SMALL LETTER E WITH GRAVE}', b'\\`e')
@@ -353,9 +371,9 @@ class LatexUnicodeTable:
         self.register(u'\N{LATIN SMALL LETTER Y WITH CIRCUMFLEX}', b'\\^y')
         self.register(u'\N{LATIN CAPITAL LETTER Y WITH DIAERESIS}', b'\\"Y')
         self.register(u'\N{LATIN CAPITAL LETTER Z WITH ACUTE}', b"\\'Z")
-        self.register(u'\N{LATIN SMALL LETTER Z WITH ACUTE}', b"\\'Z")
+        self.register(u'\N{LATIN SMALL LETTER Z WITH ACUTE}', b"\\'z")
         self.register(u'\N{LATIN CAPITAL LETTER Z WITH DOT ABOVE}', b'\\.Z')
-        self.register(u'\N{LATIN SMALL LETTER Z WITH DOT ABOVE}', b'\\.Z')
+        self.register(u'\N{LATIN SMALL LETTER Z WITH DOT ABOVE}', b'\\.z')
         self.register(u'\N{LATIN CAPITAL LETTER Z WITH CARON}', b'\\v Z')
         self.register(u'\N{LATIN SMALL LETTER Z WITH CARON}', b'\\v z')
         self.register(u'\N{LATIN CAPITAL LETTER DZ WITH CARON}', b'D\\v Z')
@@ -436,7 +454,6 @@ class LatexUnicodeTable:
         self.register(u'\N{DOT ABOVE}', b'\\.{}')
         self.register(u'\N{RING ABOVE}', b'\\r{}')
         self.register(u'\N{OGONEK}', b'\\k{}')
-        self.register(u'\N{SMALL TILDE}', b'\\~{}')
         self.register(u'\N{DOUBLE ACUTE ACCENT}', b'\\H{}')
         self.register(u'\N{LATIN SMALL LIGATURE FI}', b'fi', decode=False)
         self.register(u'\N{LATIN SMALL LIGATURE FL}', b'fl', decode=False)
@@ -475,6 +492,7 @@ class LatexUnicodeTable:
             b'\\upsilon',
             mode='math')
         self.register(u'\N{GREEK SMALL LETTER PHI}', b'\\phi', mode='math')
+        self.register(u'\N{GREEK PHI SYMBOL}', b'\\varphi', mode='math')
         self.register(u'\N{GREEK SMALL LETTER CHI}', b'\\chi', mode='math')
         self.register(u'\N{GREEK SMALL LETTER PSI}', b'\\psi', mode='math')
         self.register(u'\N{GREEK SMALL LETTER OMEGA}', b'\\omega', mode='math')
@@ -549,6 +567,16 @@ class LatexUnicodeTable:
         # match the full latex string... so decoding disabled for now
         self.register(u'Ǭ', br'\textogonekcentered{\=O}', decode=False)
         self.register(u'ǭ', br'\textogonekcentered{\=o}', decode=False)
+        self.register(u'ℕ', br'\mathbb{N}', mode='math')
+        self.register(u'ℕ', br'\mathbb N', mode='math', decode=False)
+        self.register(u'ℤ', br'\mathbb{Z}', mode='math')
+        self.register(u'ℤ', br'\mathbb Z', mode='math', decode=False)
+        self.register(u'ℚ', br'\mathbb{Q}', mode='math')
+        self.register(u'ℚ', br'\mathbb Q', mode='math', decode=False)
+        self.register(u'ℝ', br'\mathbb{R}', mode='math')
+        self.register(u'ℝ', br'\mathbb R', mode='math', decode=False)
+        self.register(u'ℂ', br'\mathbb{C}', mode='math')
+        self.register(u'ℂ', br'\mathbb C', mode='math', decode=False)
 
     def register(self, unicode_text, latex_text, mode='text', package=None,
                  decode=True, encode=True):
@@ -564,15 +592,20 @@ class LatexUnicodeTable:
         :param bool encode: Whether this translation applies to encoding
             (default: ``True``).
         """
-        if package is not None:
-            # TODO implement packages
-            pass
         if mode == 'math':
             # also register text version
             self.register(unicode_text, b'$' + latex_text + b'$', mode='text',
                           package=package, decode=decode, encode=encode)
+            self.register(unicode_text,
+                          br'\(' + latex_text + br'\)', mode='text',
+                          package=package, decode=decode, encode=encode)
             # XXX for the time being, we do not perform in-math substitutions
             return
+        if not self.lexer.binary_mode:
+            latex_text = latex_text.decode("ascii")
+        if package is not None:
+            # TODO implement packages
+            pass
         # tokenize, and register unicode translation
         self.lexer.reset()
         self.lexer.state = 'M'
@@ -585,13 +618,11 @@ class LatexUnicodeTable:
             # for instance, "\'{e}" for "\'e", "\c{c}" for "\c c", etc.
             # note: we do not remove brackets (they sometimes matter,
             # e.g. bibtex uses them to prevent lower case transformation)
-            if (len(tokens) == 2
-                and tokens[0].name.startswith('control')
-                    and tokens[1].name == 'chars'):
-                alt_tokens = (
-                    tokens[0], lexer.Token('chars', b'{'),
-                    tokens[1], lexer.Token('chars', b'}'),
-                )
+            if (len(tokens) == 2 and
+                tokens[0].name.startswith('control') and
+                    tokens[1].name == 'chars'):
+                alt_tokens = (tokens[0], self.lexer.curlylefttoken, tokens[1],
+                              self.lexer.curlyrighttoken)
                 if alt_tokens not in self.unicode_map:
                     self.max_length = max(self.max_length, len(alt_tokens))
                     self.unicode_map[alt_tokens] = u"{" + unicode_text + u"}"
@@ -599,7 +630,10 @@ class LatexUnicodeTable:
             assert len(unicode_text) == 1
             self.latex_map[unicode_text] = (latex_text, tokens)
 
+
 _LATEX_UNICODE_TABLE = LatexUnicodeTable(lexer.LatexIncrementalDecoder())
+_ULATEX_UNICODE_TABLE = LatexUnicodeTable(
+    lexer.UnicodeLatexIncrementalDecoder())
 
 # incremental encoder does not need a buffer
 # but decoder does
@@ -615,10 +649,11 @@ class LatexIncrementalEncoder(lexer.LatexIncrementalEncoder):
     """Translation table."""
 
     def __init__(self, errors='strict'):
-        lexer.LatexIncrementalEncoder.__init__(self, errors=errors)
+        super(LatexIncrementalEncoder, self).__init__(errors=errors)
         self.reset()
 
     def reset(self):
+        super(LatexIncrementalEncoder, self).reset()
         self.state = 'M'
 
     def get_space_bytes(self, bytes_):
@@ -626,14 +661,14 @@ class LatexIncrementalEncoder(lexer.LatexIncrementalEncoder):
         if self.state == 'S':
             # in space eating mode
             # control space needed?
-            if bytes_.startswith(b' '):
+            if bytes_.startswith(self.spacechar):
                 # replace by control space
-                return b'\\ ', bytes_[1:]
+                return self.controlspacechar, bytes_[1:]
             else:
                 # insert space (it is eaten, but needed for separation)
-                return b' ', bytes_
+                return self.spacechar, bytes_
         else:
-            return b'', bytes_
+            return self.emptychar, bytes_
 
     def _get_latex_bytes_tokens_from_char(self, c):
         # if ascii, try latex equivalents
@@ -649,7 +684,10 @@ class LatexIncrementalEncoder(lexer.LatexIncrementalEncoder):
         except UnicodeEncodeError:
             pass
         else:
-            return bytes_, (lexer.Token(name='chars', text=bytes_),)
+            if self.binary_mode:
+                return bytes_, (lexer.Token(name='chars', text=bytes_),)
+            else:
+                return c, (lexer.Token(name='chars', text=c),)
         # next, try latex equivalents of common unicode characters
         try:
             return self.table.latex_map[c]
@@ -663,14 +701,19 @@ class LatexIncrementalEncoder(lexer.LatexIncrementalEncoder):
                     "don't know how to translate {0} into latex"
                     .format(repr(c)))
             elif self.errors == 'ignore':
-                return b'', (lexer.Token(),)
+                return self.emptychar, (self.emptytoken,)
             elif self.errors == 'replace':
                 # use the \\char command
                 # this assumes
                 # \usepackage[T1]{fontenc}
                 # \usepackage[utf8]{inputenc}
-                bytes_ = b'{\\char' + str(ord(c)).encode("ascii") + b'}'
+                if self.binary_mode:
+                    bytes_ = b'{\\char' + str(ord(c)).encode("ascii") + b'}'
+                else:
+                    bytes_ = u'{\\char' + str(ord(c)) + u'}'
                 return bytes_, (lexer.Token(name='chars', text=bytes_),)
+            elif self.errors == 'keep' and not self.binary_mode:
+                return c,  (lexer.Token(name='chars', text=c),)
             else:
                 raise ValueError(
                     "latex codec does not support {0} errors"
@@ -737,17 +780,17 @@ class LatexIncrementalDecoder(lexer.LatexIncrementalDecoder):
                     # match!! flush buffer, and translate last bit
                     # exclude last i tokens
                     for token in self.token_buffer[:-i]:
-                        yield token.decode(self.inputenc)
+                        yield self.decode_token(token)
                     yield unicode_text
                     self.token_buffer = []
                     break
             # flush tokens that can no longer match
             while len(self.token_buffer) >= self.table.max_length:
-                yield self.token_buffer.pop(0).decode(self.inputenc)
+                yield self.decode_token(self.token_buffer.pop(0))
         # also flush the buffer at the end
         if final:
             for token in self.token_buffer:
-                yield token.decode(self.inputenc)
+                yield self.decode_token(token)
             self.token_buffer = []
 
 
@@ -772,22 +815,38 @@ class LatexCodec(codecs.Codec):
         )
 
 
+class UnicodeLatexIncrementalDecoder(LatexIncrementalDecoder):
+    table = _ULATEX_UNICODE_TABLE
+    binary_mode = False
+
+
+class UnicodeLatexIncrementalEncoder(LatexIncrementalEncoder):
+    table = _ULATEX_UNICODE_TABLE
+    binary_mode = False
+
+
 def find_latex(encoding):
     """Return a :class:`codecs.CodecInfo` instance for the requested
     LaTeX *encoding*, which must be equal to ``latex``,
     or to ``latex+<encoding>``
     where ``<encoding>`` describes another encoding.
     """
-    # check if requested codec info is for latex encoding
-    if not encoding.startswith('latex'):
+    encoding, _, inputenc_ = encoding.partition(u"+")
+    if not inputenc_:
+        inputenc_ = "ascii"
+    if encoding == "latex":
+        IncEnc = LatexIncrementalEncoder
+        DecEnc = LatexIncrementalDecoder
+    elif encoding == "ulatex":
+        IncEnc = UnicodeLatexIncrementalEncoder
+        DecEnc = UnicodeLatexIncrementalDecoder
+    else:
         return None
-    # set up all classes with correct latex input encoding
-    inputenc_ = encoding[6:] if encoding.startswith('latex+') else 'ascii'
 
-    class IncrementalEncoder_(LatexIncrementalEncoder):
+    class IncrementalEncoder_(IncEnc):
         inputenc = inputenc_
 
-    class IncrementalDecoder_(LatexIncrementalDecoder):
+    class IncrementalDecoder_(DecEnc):
         inputenc = inputenc_
 
     class Codec(LatexCodec):
@@ -803,8 +862,8 @@ def find_latex(encoding):
     return codecs.CodecInfo(
         encode=Codec().encode,
         decode=Codec().decode,
-        incrementalencoder=IncrementalEncoder_,
-        incrementaldecoder=IncrementalDecoder_,
+        incrementalencoder=Codec.IncrementalEncoder,
+        incrementaldecoder=Codec.IncrementalDecoder,
         streamreader=StreamReader,
         streamwriter=StreamWriter,
     )
