@@ -59,7 +59,7 @@ class TestDecoder(TestCase):
 
     @nose.tools.raises(TypeError)
     def test_invalid_type(self):
-        self.decode(object(), object())
+        codecs.getdecoder("latex")(object())
 
     @nose.tools.raises(ValueError)
     def test_invalid_code(self):
@@ -139,6 +139,9 @@ class TestDecoder(TestCase):
     def test_space_2(self):
         self.decode(u"æ æ", br'\ae\ \ae')
 
+    def test_space_3(self):
+        self.decode(u"æ æ", br'\ae \quad \ae')
+
     def test_number_sign_1(self):
         self.decode(u"# hello", br'\#\ hello')
 
@@ -179,6 +182,39 @@ class TestDecoder(TestCase):
         self.decode(u"\\textogonekcentered {Ō}\\textogonekcentered {ō}",
                     br'\textogonekcentered{\=O}\textogonekcentered{\=o}')
 
+    def test_math_spacing_dollar(self):
+        self.decode(u'This is a ψ test.', br'This is a $\psi$ test.')
+
+    def test_math_spacing_brace(self):
+        self.decode(u'This is a ψ test.', br'This is a \(\psi\) test.')
+
+    def test_double_math(self):
+        # currently no attempt to translate maths inside $$
+        self.decode(u'This is a $$\\psi $$ test.',
+                    br'This is a $$\psi$$ test.')
+
+    def test_tilde(self):
+        self.decode(u'This is a ˜, ˷, ∼ and ~test.',
+                    (br'This is a \~{}, \texttildelow, '
+                     br'$\sim$ and \textasciitilde test.'))
+
+    def test_backslash(self):
+        self.decode(u'This is a \\ \\test.',
+                    br'This is a $\backslash$ \textbackslash test.')
+
+    def test_percent(self):
+        self.decode(u'This is a % test.',
+                    br'This is a \% test.')
+
+    def test_math_minus(self):
+        self.decode(u'This is a − test.',
+                    br'This is a $-$ test.')
+
+    def test_swedish_again(self):
+        self.decode(
+            u"l{å}ren l{Å}ren",
+            br'l{\r a}ren l{\r A}ren')
+
 
 class TestStreamDecoder(TestDecoder):
 
@@ -224,7 +260,7 @@ class TestEncoder(TestCase):
 
     @nose.tools.raises(TypeError)
     def test_invalid_type(self):
-        self.encode(object(), object())
+        codecs.getencoder("latex")(object())
 
     # note concerning test_invalid_code_* methods:
     # u'\u2328' (0x2328 = 9000) is unicode for keyboard symbol
@@ -306,6 +342,9 @@ class TestEncoder(TestCase):
     def test_space_2(self):
         self.encode(u"æ æ", br'\ae\ \ae')
 
+    def test_space_3(self):
+        self.encode(u"æ æ", br'\ae \quad \ae')
+
     def test_number_sign(self):
         # note: no need for control space after \#
         self.encode(u"# hello", br'\# hello')
@@ -336,6 +375,30 @@ class TestEncoder(TestCase):
         self.encode(u"Ǭǭ",
                     br'\textogonekcentered{\=O}\textogonekcentered{\=o}')
 
+    def test_math_spacing(self):
+        self.encode(u'This is a ψ test.', br'This is a $\psi$ test.')
+
+    def test_double_math(self):
+        # currently no attempt to translate maths inside $$
+        self.encode(u'This is a $$\\psi$$ test.', br'This is a $$\psi$$ test.')
+
+    def test_tilde(self):
+        self.encode(u'This is a ˜, ˷, ∼ and ~test.',
+                    (br'This is a \~{}, \texttildelow , '
+                     br'$\sim$ and \textasciitilde test.'))
+
+    def test_percent(self):
+        self.encode(u'This is a % test.',
+                    br'This is a \% test.')
+
+    def test_hyphen(self):
+        self.encode(u'This is a \N{HYPHEN} test.',
+                    br'This is a - test.')
+
+    def test_math_minus(self):
+        self.encode(u'This is a − test.',
+                    br'This is a $-$ test.')
+
 
 class TestStreamEncoder(TestEncoder):
 
@@ -360,3 +423,47 @@ class TestIncrementalEncoder(TestEncoder):
             encoder.encode(text_utf8_part, final)
             for text_utf8_part, final in split_input(text_utf8))
         self.assertEqual(text_latex, b''.join(encoded_parts))
+
+
+class TestUnicodeDecoder(TestDecoder):
+
+    def decode(self, text_utf8, text_latex, inputenc=None):
+        """Main test function."""
+        text_latex = text_latex.decode(inputenc if inputenc else "ascii")
+        decoded, n = codecs.getdecoder('ulatex')(text_latex)
+        self.assertEqual((decoded, n), (text_utf8, len(text_latex)))
+
+
+class TestUnicodeEncoder(TestEncoder):
+
+    def encode(self, text_utf8, text_latex, inputenc=None, errors='strict'):
+        """Main test function."""
+        encoding = 'ulatex+' + inputenc if inputenc else 'ulatex'
+        text_latex = text_latex.decode(inputenc if inputenc else 'ascii')
+        encoded, n = codecs.getencoder(encoding)(text_utf8, errors=errors)
+        self.assertEqual((encoded, n), (text_latex, len(text_utf8)))
+
+    def uencode(self, text_utf8, text_ulatex, inputenc=None, errors='strict'):
+        """Main test function."""
+        encoding = 'ulatex+' + inputenc if inputenc else 'ulatex'
+        encoded, n = codecs.getencoder(encoding)(text_utf8, errors=errors)
+        self.assertEqual((encoded, n), (text_ulatex, len(text_utf8)))
+
+    def test_ulatex_ascii(self):
+        self.uencode(u'# ψ', u'\\# $\\psi$', 'ascii')
+
+    def test_ulatex_utf8(self):
+        self.uencode(u'# ψ', u'\\# ψ', 'utf8')
+
+    # the following tests rely on the fact that \u2328 is not in our
+    # translation table
+
+    @nose.tools.raises(ValueError)
+    def test_ulatex_ascii_invalid(self):
+        self.uencode(u'# \u2328', u'', 'ascii')
+
+    def test_ulatex_utf8_invalid(self):
+        self.uencode(u'# ψ \u2328', u'\\# ψ \u2328', 'utf8')
+
+    def test_invalid_code_keep(self):
+        self.uencode(u'# ψ \u2328', u'\\# $\\psi$ \u2328', 'ascii', 'keep')
